/*
 * BCM7315 IDE chipset-specific code
 *
 * Supports both PIO and DMA transfers, single channel (2 drives)
 *
 * Copyright (C) 2001-2003 TiVo Inc. All Rights Reserved.
 */

#include <linux/ide.h>
#include <linux/module.h>
#include <linux/plog.h>
#include <asm/irq.h>
#include <asm/bcm7315.h>
#include <asm/uaccess.h>
#include <linux/pci.h>

#include <linux/tivoconfig.h>

/*
 * m = memory, c = core, r = register, f = field, d = data.
 */
#if !defined(GET_FIELD) && !defined(SET_FIELD)
#define BCM_ALIGN(c,r,f)   c##_##r##_##f##_ALIGN
#define BCM_BITS(c,r,f)    c##_##r##_##f##_BITS
#define BCM_MASK(c,r,f)    c##_##r##_##f##_MASK
#define BCM_SHIFT(c,r,f)   c##_##r##_##f##_SHIFT

#define GET_FIELD(m,c,r,f) ((((m) & BCM_MASK(c,r,f)) >> BCM_SHIFT(c,r,f)) << BCM_ALIGN(c,r,f))

#define SET_FIELD(m,c,r,f,d) ((m) = (((m) & ~BCM_MASK(c,r,f)) | ((((d) >> BCM_ALIGN(c,r,f)) << BCM_SHIFT(c,r,f)) & BCM_MASK(c,r,f))) )

#endif /* GET & SET */

#define PRD_BYTES    8
#define PRD_ENTRIES    (PAGE_SIZE / (2 * PRD_BYTES))

#define MAX_SGLIST_SIZE (63 * 1024)

extern char * ide_xfer_verbose (uint8_t xfer_rate);

static int bcm2_dmaproc(ide_dma_action_t func, ide_drive_t *drive);
static void bcm2_resetproc(ide_drive_t *drive);

static void (*bcm2_cmd_hook)(ide_drive_t *, unsigned char *);

static unsigned int bcm2_size_priv(int dma_chunks, int block_chunks);
static void bcm2_init_drvpriv(struct idetivo_request *trq, int reading);
static int bcm2_add_dmachunk(struct idetivo_request *trq, unsigned long k_addr,
                             unsigned long u_addr, int size, int reading);
static int bcm2_add_blkchunk(struct idetivo_request *trq, int chunks, int sectors);
static int bcm2_end_blkchunk(struct idetivo_request *trq, int chunks, int sectors);
static void bcm2_set_cmd_hook(void (*handler)(ide_drive_t *, unsigned char *)); 
static void bcm2_fini_drvpriv(struct idetivo_request *trq);

static struct ll_dma_ops bcm2_sgops = {
    size_priv   :   bcm2_size_priv,
    init_drvpriv:   bcm2_init_drvpriv,
    add_dmachunk:   bcm2_add_dmachunk,
    add_blkchunk:   bcm2_add_blkchunk,
    end_blkchunk:   bcm2_end_blkchunk,
    set_cmd_hook:   bcm2_set_cmd_hook,
    fini_drvpriv:   bcm2_fini_drvpriv,
};

struct bcm2_sgent {
    u32 address;
    u32 count;
};

struct desc_chunk {
    unsigned int sglist;   //bus address of scatter list
};

enum bcm2_dma_state {
    BCM2_DMA_ERROR = 1, 
    BCM2_DMA_IDLE = 2,

    /* end */
    BCM2_DMA_END,
};

static enum bcm2_dma_state gDmaState = BCM2_DMA_IDLE;

struct bcm2_ide_priv {
    struct desc_chunk *first_chunk;
    struct desc_chunk *cur_chunk;
    struct bcm2_sgent *first_sgent;
    struct bcm2_sgent *cur_sgent;
    int big_flush;
    int last_size;
    u32 bus_addr_end;
    int sgmem_nfrags;
    struct bcm2_sgent **sgmem_frags;
    struct bcm2_sgent *sgmem_scratch_frag;
    int sgmem_scratch_size;
};

#ifdef BCM_IDE_DEBUG
static inline void print_sg_list(struct bcm2_ide_priv *priv);
#endif /* BCM_IDE_DEBUG */


static int bcm2_ide_default_irq(ide_ioreg_t base)
{
    return BCM_LINUX_IDE0_IRQ;
}

// Rather than the actual IO base (since we only use 1 IDE interface),
// this routine just returns 1 to indicate the interface is valid, or 0 to
// indicate it is not.  That value gets passed to init_hwif_ports, below.
static ide_ioreg_t bcm2_ide_default_io_base(int index)
{
    if (index)
        return 0;
    else
        return 1;
}

static void bcm2_ide_init_hwif_ports(hw_regs_t *hw, ide_ioreg_t data_port,
                         ide_ioreg_t ctrl_port, int *irq)
{
    int i;
    TivoConfigValue ide_cap;
    unsigned int chip_id;
    if (!data_port)
        return;

    chip_id = (*BCM_CHIPID >> 16);
    printk("[%s] : chip ID = 0x%x\n", __PRETTY_FUNCTION__, chip_id);


    *BCM_IDE_CMD = 0x0005;
    *BCM_IDE_PROG_IF = 0x8F00;
    *BCM_IDE_PRI_CS0_ADDR = 0x200;
    *BCM_IDE_PRI_CS1_ADDR = 0x240;
    *BCM_IDE_SEC_CS0_ADDR = 0x280;
    *BCM_IDE_SEC_CS1_ADDR = 0x2C0;
    *BCM_IDE_BMIDE_ADDR = 0x300;
    *BCM_IDE_PIO_TIMING = 0x99999999;
    *BCM_IDE_DMA_TIMING = 0xFFFFFFFF;
    *BCM_IDE_PIO_CONTROL = 0xF0F0;
    *BCM_IDE_UDMA_CONTROL = 0;
    *BCM_IDE_UDMA_MODE = 0x040F;
#ifdef CONFIG_TIVO_CPU_TO_BE32
    *BCM_IDE_IF_CONTROL = 0x00100012;
#else
    *BCM_IDE_IF_CONTROL = 0x00120012;
#endif /* CONFIG_TIVO_CPU_TO_BE32 */
    *BCM_IDE_MBIST_CONTROL = 0x22;
    *BCM_IDE_INT_MASK = 0x22;
    *BCM_IDE_INT_STATUS = 0x22;
    *BCM_IDE_PRI_BM_CMD_STAT = 0x40000;
    *BCM_IDE_PRI_BM_DESC_TAB = 0;
    *BCM_IDE_SEC_BM_CMD_STAT = 0x40000;
    *BCM_IDE_SEC_BM_DESC_TAB = 0;

    for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
        hw->io_ports[i] =
            (ide_ioreg_t)((unsigned long)BCM_IDE_PRI_DATA + i);
    }
    hw->io_ports[IDE_CONTROL_OFFSET] = (ide_ioreg_t)BCM_IDE_PRI_CMD_STAT;

    if (irq != NULL) {
        *irq = BCM_LINUX_IDE0_IRQ;
    }
    hw->io_ports[IDE_IRQ_OFFSET] = BCM_LINUX_IDE0_IRQ;


    if (GetTivoConfig(kTivoConfigIdeCap, &ide_cap)) {
        ide_hwifs[0].udma_four = 0;
    } else {
        ide_cap &= kTivoConfigIdeCapAta66;
        ide_hwifs[0].udma_four = ide_cap ? 1 : 0;
    }
    ide_setup_dma(&ide_hwifs[0], (unsigned long) BCM_IDE_PRI_BM_CMD_STAT,8);
    ide_hwifs[0].dmaproc = bcm2_dmaproc;
    ide_hwifs[0].resetproc = bcm2_resetproc;

    ide_hwifs[0].tivo.version = IDETIVO_DRIVER_VERSION;
    ide_hwifs[0].tivo.sgops = &bcm2_sgops;
}

static int bcm2_ide_request_irq(unsigned int irq,
                                void (*handler)(int,void *, struct pt_regs *),
                                unsigned long flags, const char *device,
                                void *dev_id)
{
    // This is always BCM_LINUX_IDE0_IRQ
    return request_irq(irq, handler, flags, device, dev_id);
}

static void bcm2_ide_free_irq(unsigned int irq, void *dev_id)
{
    free_irq(irq, dev_id);
}

static int bcm2_ide_check_region(ide_ioreg_t from, unsigned int extent)
{
    return 0;
}

static void bcm2_ide_request_region(ide_ioreg_t from, unsigned int extent,
                                    const char *name)
{
    return;
}

static void bcm2_ide_release_region(ide_ioreg_t from, unsigned int extent)
{
    return;
}

struct ide_ops bcm_ide_ops = {
    &bcm2_ide_default_irq,
    &bcm2_ide_default_io_base,
    &bcm2_ide_init_hwif_ports,
    &bcm2_ide_request_irq,
    &bcm2_ide_free_irq,
    &bcm2_ide_check_region,
    &bcm2_ide_request_region,
    &bcm2_ide_release_region
};

void _tivo_ide_out_byte(u8 data, ide_ioreg_t port)
{
    *(volatile u8 *)port = data;
}
EXPORT_SYMBOL(_tivo_ide_out_byte);

u8 _tivo_ide_in_byte(ide_ioreg_t port)
{
    return *(volatile u8 *)port;
}
EXPORT_SYMBOL(_tivo_ide_in_byte);

void _tivo_ide_out_word(u16 data, ide_ioreg_t port)
{
	*(volatile u16 *)port = ((data >> 8) & 0xff) | (data<<8);
}

/* This is only ever used for data, rest of regs are 8-bit... */
u16 _tivo_ide_in_word(ide_ioreg_t port)
{
    return *BCM_IDE_PRI_DATA;
}

// The following are replacements for the ones removed from drivers/ide/ide.c

/*
 * This is used for most PIO data transfers *from* the IDE interface
 */
void ide_input_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
    u16 *ptr = (u16 *)buffer;

    while (wcount--) {
        barrier();
        *ptr++ = *BCM_IDE_PRI_DATA;
        barrier();
        *ptr++ = *BCM_IDE_PRI_DATA;
    }
}

/*
 * This is used for most PIO data transfers *to* the IDE interface
 */
void ide_output_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
    u16 *ptr = (u16 *)buffer;

    while (wcount--) {
        *BCM_IDE_PRI_DATA = *ptr++;
        barrier();
        *BCM_IDE_PRI_DATA = *ptr++;
        barrier();
    }
}

/*
 * The following routines are mainly used by the ATAPI drivers.
 *
 * These routines will round up any request for an odd number of bytes,
 * so if an odd bytecount is specified, be sure that there's at least one
 * extra byte allocated for the buffer.
 */
void atapi_input_bytes (ide_drive_t *drive, void *buffer, unsigned int bytecount)
{
    unsigned short *ptr = (unsigned short *) buffer;
    unsigned int wcount;

    ++bytecount;
    wcount = (bytecount / 4) << 1;

    while (wcount--) {
        *ptr++ = *BCM_IDE_PRI_DATA;
        barrier();
    }
    if ((bytecount & 0x03) >= 2) {
        *ptr++ = *BCM_IDE_PRI_DATA;
        barrier();
    }
}

void atapi_output_bytes (ide_drive_t *drive, void *buffer, unsigned int bytecount)
{
    unsigned short *ptr = (unsigned short *) buffer;
    unsigned int wcount;

    ++bytecount;
    wcount = (bytecount / 4) << 1;

    while (wcount--) {
        *BCM_IDE_PRI_DATA = *ptr++;
        barrier();
    }
    if ((bytecount & 0x03) >= 2) {
        *BCM_IDE_PRI_DATA = *ptr++;
        barrier();
    }
}

static inline byte get_fastest_speed(ide_drive_t *drive) 
{
    struct hd_driveid *id = drive->id;
    uint ultra = 0;
    uint mword = 0;
    uint8_t speed = 0;
    uint8_t ultra66 = 0;

    if (id->field_valid & 4) {
        /* is ultra dma supported */
        ultra = id->dma_ultra;
        /* check if 80-pin cable is attached */
        ultra66 = (uint8_t)(HWIF(drive)->udma_four);
#ifdef DEBUG
        if (ultra66 == 0)
            printk("%s: No 80-pin cable attached\n", HWIF(drive)->name);
#endif /* DEBUG */
    }
    if (id->field_valid & 2) {
        /* multi-word DMA supported? */
        mword = id->dma_mword;
    }

    if ((ultra & 0x0010) && (ultra66)) {
        speed = XFER_UDMA_4;
    } else if ((ultra & 0x0008) && (ultra66)) {
        speed = XFER_UDMA_3;
    } else if (ultra & 0x0004) {
        speed = XFER_UDMA_2;
    } else if (ultra & 0x0002) {
        speed = XFER_UDMA_1;
    } else if (ultra & 0x0001) {
        speed = XFER_UDMA_0;
    } else if (mword & 0x0004) {
        speed = XFER_MW_DMA_2;
    } else if (mword & 0x0002) {
        speed = XFER_MW_DMA_1;
    } else if (mword & 0x0001) {
        speed = XFER_MW_DMA_0;
    }

    return speed;
}


static int tune_ide_chipset(ide_drive_t * drive, byte speed)
{
    /* unit is the drive select number, 0 or 1 */
    u8 unit          = (drive->select.b.unit & 0x01);
    u32 regTimingPIO   = 0x0;
    u32 regTimingDMA   = 0x0;
    u32 regCtlPIO      = 0x0;
    u32 regCtlStatUDMA = 0x0;
    int ret            = 0;
    int err            = 0;
    /* 
        Actual time is computed as (xxXXX + 1 ) * (1/33MHz) 
        example cwPio = 3  ==>  (3 + 1) * 30ns = 120ns
    */
    u8 cwPIO, cwDMA;     /* command width for PIO and DMA */
    u8 modePIO;          /* PIO mode */
    u8 rwPIO, rwDMA;     /* recover width for PIO and DMA */
    u8 enUDMA;           /* UDMA enable for primary channel */
    u8 modeUDMA;         /* UDMA mode */

    /* read the registers responsible for timing in PIO and DMA modes */
    regTimingPIO   = *BCM_IDE_PIO_TIMING;
    regTimingDMA   = *BCM_IDE_DMA_TIMING;
    regCtlPIO      = *BCM_IDE_PIO_CONTROL;
    regCtlStatUDMA = *BCM_IDE_UDMA_CONTROL;

    cwPIO = unit ? GET_FIELD(regTimingPIO, IDE, PIO_TIME, PS_PIO_CMD) :
        GET_FIELD(regTimingPIO, IDE, PIO_TIME, PM_PIO_CMD); 

    rwPIO = unit ? GET_FIELD(regTimingPIO, IDE, PIO_TIME, PS_PIO_CMD_REC) :
        GET_FIELD(regTimingPIO, IDE, PIO_TIME, PM_PIO_CMD_REC);

    modePIO = unit ? GET_FIELD(regCtlPIO, IDE, PIO_CNTL_MODE, PS_PIO_ACC_MD) :
        GET_FIELD(regCtlPIO, IDE, PIO_CNTL_MODE, PM_PIO_ACC_MD);

    /* retrieve current settings for command and command recovery timing for DMA */
    cwDMA = unit ? GET_FIELD(regTimingDMA, IDE, DMA_TIME, PS_DMA_CMD) :
        GET_FIELD(regTimingDMA, IDE, DMA_TIME, PM_DMA_CMD);
    rwDMA = unit ? GET_FIELD(regTimingDMA, IDE, DMA_TIME, PS_DMA_CMD_REC) :
        GET_FIELD(regTimingDMA, IDE, DMA_TIME, PM_DMA_CMD_REC);

    /* retrieve current settings for UDMA.  The mode defines timing for UDMA */
    enUDMA = unit ? GET_FIELD(regCtlStatUDMA, IDE, UDMA_CNTL_STAT, PS_UDMA) :
        GET_FIELD(regCtlStatUDMA, IDE, UDMA_CNTL_STAT, PM_UDMA);
    modeUDMA = unit ? GET_FIELD(regCtlStatUDMA, IDE, UDMA_CNTL_STAT, PS_UDMA_MODE) :
        GET_FIELD(regCtlStatUDMA, IDE, UDMA_CNTL_STAT, PM_UDMA_MODE);

    switch (speed) {
        case XFER_UDMA_4:
        case XFER_UDMA_3:
        case XFER_UDMA_2:
        case XFER_UDMA_1:
        case XFER_UDMA_0:
            enUDMA = IDE_UDMA_CNTL_STAT_SS_UDMA_ENABLE;  /* pri/sec master/slave macro are the same */
            modeUDMA = speed & 0xF;
            cwDMA = 0x2; rwDMA = 0x0;                    /* can do mdma2 too */
            cwPIO = 0x2; rwPIO = 0x0; modePIO = 0x4;     /* can do pio4 too */
            break;

        case XFER_MW_DMA_2:
            cwDMA = 0x2; rwDMA = 0x0;
            enUDMA = IDE_UDMA_CNTL_STAT_SS_UDMA_DISABLE; /* can't do udma */
            cwPIO = 0x2; rwPIO = 0x0; modePIO = 0x4;     /* can do pio4 too */
            break;   
        case XFER_MW_DMA_1:
            cwDMA = 0x2; rwDMA = 0x1;
            enUDMA = IDE_UDMA_CNTL_STAT_SS_UDMA_DISABLE; /* can't do udma */
            cwPIO = 0x2; rwPIO = 0x0; modePIO = 0x4;     /* can do pio4 too */
            break;   
        case XFER_MW_DMA_0:
            cwDMA = 0x7; rwDMA = 0x7;
            enUDMA = IDE_UDMA_CNTL_STAT_SS_UDMA_DISABLE; /* can't do udma */
            cwPIO = 0x2; rwPIO = 0x0; modePIO = 0x4;     /* can do pio4 too */
            break;   

        /* assumption is that we don't support Single-Word DMA */
        case XFER_SW_DMA_2:
        case XFER_SW_DMA_1:
        case XFER_SW_DMA_0:
            printk("%s: single word DMA not supported\n", drive->name);
            return 0;

        case XFER_PIO_4:
            cwPIO = 0x2; rwPIO = 0x0; modePIO = 0x4;
            enUDMA = IDE_UDMA_CNTL_STAT_SS_UDMA_DISABLE; /* can't do udma */
            break;
        case XFER_PIO_3:
            cwPIO = 0x2; rwPIO = 0x2; modePIO = 0x3;
            enUDMA = IDE_UDMA_CNTL_STAT_SS_UDMA_DISABLE; /* can't do udma */
            break;
        case XFER_PIO_2:
            cwPIO = 0x3; rwPIO = 0x4; modePIO = 0x2;
            enUDMA = IDE_UDMA_CNTL_STAT_SS_UDMA_DISABLE; /* can't do udma */
            break;
        case XFER_PIO_1:
            cwPIO = 0x4; rwPIO = 0x7; modePIO = 0x1;
            enUDMA = IDE_UDMA_CNTL_STAT_SS_UDMA_DISABLE; /* can't do udma */
            break;
        case XFER_PIO_0:
            cwPIO = 0x5; rwPIO = 0xD; modePIO = 0x0;
            enUDMA = IDE_UDMA_CNTL_STAT_SS_UDMA_DISABLE; /* can't do udma */
            break;

        default:
            ret = 1;
    }
    if (ret) {
        return ret;
    }

    /* modify the settings for command and command recovery timing for DMA */
    unit ? SET_FIELD(regTimingPIO, IDE, PIO_TIME, PS_PIO_CMD, cwPIO) :
        SET_FIELD(regTimingPIO, IDE, PIO_TIME, PM_PIO_CMD, cwPIO);

    unit ? SET_FIELD(regTimingPIO, IDE, PIO_TIME, PS_PIO_CMD_REC, rwPIO) :
        SET_FIELD(regTimingPIO, IDE, PIO_TIME, PM_PIO_CMD_REC, rwPIO);

    unit ? SET_FIELD(regCtlPIO, IDE, PIO_CNTL_MODE, PS_PIO_ACC_MD, modePIO) :
        SET_FIELD(regCtlPIO, IDE, PIO_CNTL_MODE, PM_PIO_ACC_MD, modePIO);

    /* modify settings for command and command recovery timing for DMA */
    unit ? SET_FIELD(regTimingDMA, IDE, DMA_TIME, PS_DMA_CMD, cwDMA) :
        SET_FIELD(regTimingDMA, IDE, DMA_TIME, PM_DMA_CMD, cwDMA);
    unit ? SET_FIELD(regTimingDMA, IDE, DMA_TIME, PS_DMA_CMD_REC, rwDMA) :
        SET_FIELD(regTimingDMA, IDE, DMA_TIME, PM_DMA_CMD_REC, rwDMA);

    /* modify settings for UDMA.  The mode defines timing for UDMA */
    unit ? SET_FIELD(regCtlStatUDMA, IDE, UDMA_CNTL_STAT, PS_UDMA, enUDMA) :
        SET_FIELD(regCtlStatUDMA, IDE, UDMA_CNTL_STAT, PM_UDMA, enUDMA);
    unit ? SET_FIELD(regCtlStatUDMA, IDE, UDMA_CNTL_STAT, PS_UDMA_MODE, modeUDMA) :
        SET_FIELD(regCtlStatUDMA, IDE, UDMA_CNTL_STAT, PM_UDMA_MODE, modeUDMA);

    /* write the registers we just modified above */
    *BCM_IDE_PIO_TIMING   = regTimingPIO; 
    *BCM_IDE_DMA_TIMING   = regTimingDMA; 
    *BCM_IDE_PIO_CONTROL  = regCtlPIO;
    *BCM_IDE_UDMA_CONTROL = regCtlStatUDMA;
    barrier();

    if (speed > XFER_PIO_4) {
        // Now that the chipset is configured, do the drive
        err = ide_config_drive_speed(drive, speed);
        drive->current_speed = speed;
    } 

#ifdef DEBUG
    printk("%s: Configured for %s %s\n", 
            drive->name, ide_xfer_verbose(speed), err ? "[Error]" : "");
#endif /* DEBUG */
    return 0;
}



static inline int configure_drive_for_dma(ide_drive_t *drive)
{
    struct hd_driveid *id = drive->id;
    byte speed;
    ide_hwif_t *hwif = HWIF(drive);

    hwif->autodma = 1;
    drive->present = 1;

    if (id && (id->capability & 1) && hwif->autodma) {
        /* consult the list of known "bad" drives" */
        if (hwif->dmaproc(ide_dma_bad_drive, drive)) {
            return hwif->dmaproc(ide_dma_off, drive);
        }

        speed = get_fastest_speed(drive);
        tune_ide_chipset(drive, speed);

        if (!drive->init_speed) {
            drive->init_speed = speed;
        }

#ifdef DEBUG
        byte dma_stat = inb(ide_hwifs[0].dma_base+2);
        printk("\nBIOS settings: %s:%s, %s:%s",
               hwif->drives[0].name, (dma_stat & 0x20) ? "DMA" : "pio",
               hwif->drives[1].name, (dma_stat & 0x40) ? "DMA" : "pio");
#endif /* DEBUG */

        if (hwif->dmaproc(ide_dma_good_drive, drive))
            return hwif->dmaproc(ide_dma_on, drive);

        if (speed > 0) {
            drive->using_dma = 1;
            return hwif->dmaproc(ide_dma_on, drive);
        }

        return hwif->dmaproc(ide_dma_off_quietly, drive);
    }
}

/*
 * over-rides the default ide_dmaproc in ide_dma.c
 * Returns 0 if all went well.
 * Returns 1 if DMA read/write could not be started, in which case
 * the caller should revert to PIO for the current request.
 */
static int bcm2_dmaproc(ide_dma_action_t func, ide_drive_t *drive)
{
    switch (func) {
        case ide_dma_check:
            return configure_drive_for_dma(drive);
        case ide_dma_read:
        case ide_dma_write:
            *__IDEREG32(0x9c) = 1;
            barrier();
            *__IDEREG32(0xc0) = 0;
            *__IDEREG32(0xc4) = 0;
            *__IDEREG32(0xc8) = 0;
            *__IDEREG32(0xcc) = 0;
            break;
        case ide_dma_test_irq: /* returns 1 if dma irq issued, 0 otherwise */
            return 1;
        case ide_dma_lostirq:
            gDmaState = BCM2_DMA_ERROR;
            return 1;
        case ide_dma_end:
            {
                enum bcm2_dma_state oldDmaState = gDmaState;
                int ret = 0;
                gDmaState = BCM2_DMA_IDLE;
                ret = ide_dmaproc(func, drive);
                if (0 == ret) {
                    if (oldDmaState == BCM2_DMA_ERROR) {
                        return 1;
                    }
                }
                return ret;
            }

        default:
            break;
    }

    return ide_dmaproc(func, drive);
}


static void bcm2_resetproc(ide_drive_t *drive) 
{
    ide_hwif_t *hwif = HWIF(drive);
    unsigned int unit;

    // in some cases, the ide reset code will turn off DMA for a drive
    // other than the one that caused the reset, so recheck all drives
    // on this hwif
    for (unit = 0; unit < MAX_DRIVES; ++unit) {
        ide_drive_t *check_drv = &hwif->drives[unit];

        if (check_drv->present && !check_drv->using_dma) {
            bcm2_dmaproc(ide_dma_check, check_drv);
            printk("%s: re-enabled DMA\n", check_drv->name);
        }
    }
    return;
}

/*
 * Memory allocation - pretty much for sglists
 * The returned memory is cache-line aligned
 */

// Start of memory returned by kmalloc
// Cache alignment space
struct bcm2_mem_free {
    void *kmalloc_addr;
    unsigned int max_size;
    struct bcm2_mem_free *prev;
    struct bcm2_mem_free *next;
};
// start of memory returned to caller

static struct bcm2_mem_free *bcm2_mem_freelist = NULL;
static struct {
    int size;
    int slop;
    int scratch;
    int kmallocs;
    int nscratch;
} bcm2_alloc_stats;

static void *bcm2_alloc_priv(unsigned int size)
{
    struct bcm2_mem_free *candidate, *best;
    void *kmalloc_addr;
    void *ret_addr;
    unsigned int needed;

    spin_lock(&io_request_lock);
    for (best = bcm2_mem_freelist ; best ; best = best->next) {
        if (best->max_size >= size) {
            break;
        }
    }
    /*
      If we found one, unlink it.
    */
    if (best) {
        if (best->next) {
	        best->next->prev = best->prev;
        }
        if (best->prev) {
	        best->prev->next = best->next;
        } else {
	        bcm2_mem_freelist = best->next;
        }
      spin_unlock(&io_request_lock);
      return (((char *)best) + sizeof(struct bcm2_mem_free));
    }
    spin_unlock(&io_request_lock);
    
    /* No suitable request found.  Allocate a new one. 
     * Make room for table (cache-line aligned) and roundup to pagesize 
     */
    needed = PAGE_SIZE * ((sizeof (struct bcm2_mem_free) + size + L1_CACHE_BYTES - 4 + PAGE_SIZE - 1) / PAGE_SIZE);
    kmalloc_addr = (struct bcm2_mem_free *) kmalloc(needed, GFP_KERNEL);
    if (!kmalloc_addr) {
        return NULL;  /* This really should never happen! */
    }
    ret_addr = (void *) L1_CACHE_ALIGN((unsigned long)kmalloc_addr + sizeof(struct bcm2_mem_free));
    candidate = ret_addr - sizeof(struct bcm2_mem_free);
    candidate->kmalloc_addr = kmalloc_addr;
    candidate->max_size = needed - (ret_addr - kmalloc_addr);
    bcm2_alloc_stats.size += needed;
    bcm2_alloc_stats.slop += needed - size;
    bcm2_alloc_stats.kmallocs++;
    return ret_addr;
}

static void bcm2_release_priv(void *addr)
{
    struct bcm2_mem_free *req, *next, *prev = NULL;


    req = (struct bcm2_mem_free *) (addr - sizeof(struct bcm2_mem_free));
    spin_lock(&io_request_lock);
    next = bcm2_mem_freelist;
    while (next && next->max_size < req->max_size) {
        prev = next;
        next = next->next;
    }
    req->next = next;
    req->prev = prev;
    if (next) {
        next->prev = req;
    }
    if (prev) {
        prev->next = req;
    } else {
        bcm2_mem_freelist = req;
    }
    spin_unlock(&io_request_lock);
}

/*
 These functions are used to preprocess TiVo I/O requests

 Report back how much room needed to allocate along with the request
 for use by the DMA engine, based on number of scatter list entries,
 and number of distinct scatter lists.
 N.B.: we only return size needed for the priv struct
 + number of ptrs to lists
 + number of list fragment pointers (1 frag/list)
 Space for the sg list entries themselves is allocated in add_blkchunk()
 Each separate IDE transfer
 (of max MAX_NRSECTORS sectors) requires a separate scatter list.

 Structure of the idetivo request + extensions (allocated in
 ide-disk.c:allocate_request())
    idetivo_request
    fsIovec
    pagelist (nchunks+1) * 4
    ** bcm2_ide_priv
    ** chunklist (desc_chunk array of hardware ptrs to sglists)
    ** sgmem_frags (array of ptrs to memory frags containing hardware sglists

 ** = size returned by bcm2_size_priv
*/
static unsigned int bcm2_size_priv(int dma_chunks, int block_chunks)
{
    unsigned int size = sizeof(struct bcm2_ide_priv) +
               block_chunks * sizeof(struct desc_chunk) +
               block_chunks * sizeof(struct bcm2_sgent *);
    return size;
}

/*
  put the chunk list right after the priv struct
  put the scatter list after the chunk list (but cache line align it)
*/
static void bcm2_init_drvpriv(struct idetivo_request *trq, int reading)
{
    struct bcm2_ide_priv *priv = trq->drv_priv;
    int nsglists = trq->num_chunks;

    // this is a bit of platform specificness, but there's no generic way
    // to do it: if the request totals more than twice the dcache size,
    // mark to do one big flush, instead of flushing each individual chunk
 
#ifdef CONFIG_TGC
    priv->big_flush = (trq->req->nr_sectors > 32 * 1024 / SECTOR_BYTES);
#else
    priv->big_flush = (trq->req->nr_sectors > 16 * 1024 / SECTOR_BYTES);
#endif

    priv->bus_addr_end = 1; // an impossible addr

    priv->first_chunk = (struct desc_chunk *)((unsigned long)priv +
                                              sizeof(struct bcm2_ide_priv));
    priv->cur_chunk = priv->first_chunk;

    priv->sgmem_frags = (struct bcm2_sgent **) (priv->first_chunk + nsglists);
    priv->sgmem_nfrags = 0;
    memset(priv->sgmem_frags, 0, nsglists * sizeof(struct bcm2_sgent *));

    /* Put sgents in scratch frag - move to real one when we know list size */
    priv->sgmem_scratch_size = 2 * PAGE_SIZE - 1024; 
    priv->sgmem_scratch_frag = 
        (struct bcm2_sgent *) bcm2_alloc_priv(priv->sgmem_scratch_size);
    priv->first_sgent = priv->sgmem_scratch_frag;
    priv->cur_sgent = priv->first_sgent;
    bcm2_alloc_stats.scratch = priv->sgmem_scratch_size;
}

/*
    Add a chunk to the DMA scatter list and do any necessary cache flushing
    Both kernel and user virtual addresses are passed in, in case the user
    address is needed for flushing purposes (to avoid virtual aliasing
    problem on CPUs with virually indexed caches).  The user address is
    guaranteed not to page fault and size is guaranteed to be non-zero.
*/
static int bcm2_add_dmachunk(struct idetivo_request *trq, unsigned long k_addr,
                             unsigned long u_addr, int size, int reading)
{
    struct bcm2_ide_priv *priv = trq->drv_priv;
    struct bcm2_sgent *sgent = priv->cur_sgent;
    u32 bus_addr;
    int total_size;
    int cur_sgsize;

    if ((k_addr | size) & 3) {
        printk(KERN_WARNING "bcm2_ide: DMA not word aligned\n");
        return -EINVAL;
    }

    if (!priv->big_flush) {
        if (reading) {
            dma_cache_inv(u_addr, size);
        } else {
            dma_cache_wback_inv(u_addr, size);
        }
    }

    bus_addr = (u32)virt_to_bus((void *)k_addr);
    total_size = size;
    u32 cur_len = total_size;

    while (cur_len) {
        u32 xcount, bcount;

        // If sglist overflows scratch space, realloc larger area
        cur_sgsize = (sgent - priv->first_sgent) * sizeof(struct bcm2_sgent);
        if (cur_sgsize > MAX_SGLIST_SIZE) {
            printk("%s: DMA LIST TOO LONG (> %d)!: FAILING REQUEST\n", 
                    __FUNCTION__, MAX_SGLIST_SIZE);
            panic("DMA LIST TOO LONG");
            // No return
        }
        if (cur_sgsize > priv->sgmem_scratch_size) {
            struct bcm2_sgent *old_scratch = priv->sgmem_scratch_frag;
            priv->sgmem_scratch_size = 
                2 * (priv->sgmem_scratch_size + 1024) - 1024;
            priv->sgmem_scratch_frag =
              (struct bcm2_sgent *)bcm2_alloc_priv(priv->sgmem_scratch_size);
            memcpy(priv->sgmem_scratch_frag, old_scratch, cur_sgsize);
            bcm2_release_priv(old_scratch);
            priv->cur_sgent = priv->sgmem_scratch_frag + 
                                (priv->cur_sgent - priv->first_sgent);
            priv->first_sgent = priv->sgmem_scratch_frag;
            bcm2_alloc_stats.scratch = priv->sgmem_scratch_size;
            bcm2_alloc_stats.nscratch++;
            sgent = priv->cur_sgent;
        }
        
        bcount = 0x10000 - (bus_addr & 0xffff);
        if (bcount > cur_len)
            bcount = cur_len;

        xcount = bcount & 0xffff;

#ifdef CONFIG_TIVO_CPU_TO_BE32
        sgent->address = cpu_to_be32(bus_addr);
        sgent->count = cpu_to_be32(xcount);
#else
        sgent->address = cpu_to_le32(bus_addr);
        sgent->count = cpu_to_le32(xcount);
#endif /* CONFIG_TIVO_CPU_TO_BE32 */
        bus_addr += bcount;
        cur_len -= bcount;

        priv->cur_sgent = sgent + 1;
        sgent = priv->cur_sgent;
    }
    priv->last_size = total_size;
    priv->bus_addr_end = bus_addr + size;
    return 0;
}

// Finish up a block of DMA chunks that makes up a single IDE transfer,
// and start a new one
//
static int bcm2_add_blkchunk(struct idetivo_request *trq, int chunks, int sectors)
{
    struct bcm2_ide_priv *priv = trq->drv_priv;
    struct desc_chunk *desc = priv->cur_chunk++;
    int sglist_size;
    struct bcm2_sgent *sglist;

#ifdef CONFIG_TIVO_CPU_TO_BE32
    (priv->cur_sgent - 1)->count |= cpu_to_be32(0x80000000);
#else
    (priv->cur_sgent - 1)->count |= cpu_to_le32(0x80000000);
#endif

    // Now that list size is known, create and fill real list from scratch one
    sglist_size = 
        (priv->cur_sgent - priv->first_sgent) * sizeof(struct bcm2_sgent);
    sglist = bcm2_alloc_priv(sglist_size);
    if (sglist == NULL) {
        return -ENOMEM;
    }
    memcpy(sglist, priv->sgmem_scratch_frag, sglist_size);
    priv->sgmem_frags[priv->sgmem_nfrags] = sglist;
    priv->sgmem_nfrags++;
    if (priv->sgmem_nfrags > trq->num_chunks + 1) {
        // Debugging check - This cannot happen
        printk("%s: USED UP ALL FRAG LIST PTRS\n", __FUNCTION__);
    } 

    // Reset the scratch sglist buffer for next sglist to use
    priv->cur_sgent = priv->first_sgent;

    // Setup hardware list ptr to real sglist
    desc->sglist = virt_to_phys(sglist);

    priv->bus_addr_end = 1; // an impossible addr

    // Flush real sglist: if flushing whole cache, will flush after last sglist
    if (!priv->big_flush) {
        dma_cache_wback_inv((unsigned long)sglist, (unsigned long)sglist_size);
    }
    return 0;
}

// Finish up the last block of DMA chunks, and do any final prep on the
// data going to the DMA engine
//
static int bcm2_end_blkchunk(struct idetivo_request *trq, int chunks, int sectors)
{
    struct bcm2_ide_priv *priv = trq->drv_priv;
    int ret = 0;

    if (priv->bus_addr_end == 1) {
        // Sanity check - can't happen
        printk("%s: already added last blk!\n", __FUNCTION__);
        panic("already added last blk");
    }

    // Finish the last block
    ret = bcm2_add_blkchunk(trq, chunks, sectors);
    if (ret)
        goto out;

    // If not flushing whole cache, already flushed each sglist individually
    if (priv->big_flush) {
        flush_dcache();
    }
    priv->cur_chunk = priv->first_chunk;
out:
    bcm2_release_priv(priv->sgmem_scratch_frag);
    priv->sgmem_scratch_frag = NULL;
    return ret;
}

static void bcm2_set_cmd_hook(void (*handler)(ide_drive_t *, unsigned char *)) {
    bcm2_cmd_hook = handler;
}

static void bcm2_fini_drvpriv(struct idetivo_request *trq)
{
    struct bcm2_ide_priv *priv = trq->drv_priv;
    int i;

    for (i=0; i < priv->sgmem_nfrags; i++) {
        if (priv->sgmem_frags[i] != NULL) {
            bcm2_release_priv(priv->sgmem_frags[i]);
        }
    }
}

ide_startstop_t bcm2_dma_intr(ide_drive_t *drive)
{
    int i, dma_stat;
    byte stat;
    struct request *rq = HWGROUP(drive)->rq;
    struct idetivo_request *trq = rq->tivo;

    dma_stat = bcm2_dmaproc(ide_dma_end, drive);
    stat = GET_STAT();            /* get drive status */
    if (OK_STAT(stat,DRIVE_READY,drive->bad_wstat|DRQ_STAT)) {
        if (!dma_stat) {
            plog_diskstop(rq);
            if (trq) {
                unsigned int max_nrsectors = trq->max_nrsectors;

                if (rq->nr_sectors > max_nrsectors) {
                    ide_hwif_t *hwif = HWIF(drive);
                    ide_startstop_t startstop;

                    rq->nr_sectors -= max_nrsectors;
                    rq->sector += max_nrsectors;
                    trq->progress += max_nrsectors * SECTOR_BYTES;
                    add_blkdev_randomness(MAJOR(rq->rq_dev));
                    SELECT_DRIVE(hwif, drive);
                    if (ide_wait_stat(&startstop, drive, drive->ready_stat,
                                      BUSY_STAT|DRQ_STAT, WAIT_READY)) {
                        printk("%s: drive not ready for command\n", drive->name);
                        return startstop;
                    }
                    trq->scramble.recordOffset++;
                    return DRIVER(drive)->do_request(drive, rq, rq->sector
                                                      + trq->partition_start);
                } else {
                    trq->progress += rq->nr_sectors * SECTOR_BYTES;
                ide_end_request(1, HWGROUP(drive));
                trq->status = IDETIVO_DONE;
                return ide_stopped;
                }
            } else {
                for (i = rq->nr_sectors; i > 0;) {
                    i -= rq->current_nr_sectors;
                    ide_end_request(1, HWGROUP(drive));
                }
            }
            return ide_stopped;
        }
        printk("%s: dma_intr: bad DMA status\n", drive->name);
    }
    if (trq) {
        struct bcm2_ide_priv *priv = trq->drv_priv;

        // if the request gets retried, we need to reprocess this
        // chunk of the DMA scatterlist
        priv->cur_chunk--;
        trq->status = IDETIVO_ERROR;
        trq->drive_status = stat;
        // XXX: should probably read drive regs to query how far we
        // got in last 256 sector (max) chunk, too...
    }
    return ide_error(drive, "dma_intr", stat);
}

int bcm2_build_dmatable(ide_drive_t *drive) 
{
    struct request *rq = HWGROUP(drive)->rq;
    
    if (!rq->tivo) {
        return 1;
    }
    struct bcm2_ide_priv *priv = rq->tivo->drv_priv;
    struct desc_chunk *chunk = priv->cur_chunk++;
    struct io_scramble *pscramble = &rq->tivo->scramble;
    ide_hwif_t *hwif = HWIF(drive);

    rq->tivo->status = IDETIVO_INPROGRESS;

    if (pscramble->magic == IO_SCRAMBLE_MAGIC &&
        pscramble->version == IO_SCRAMBLE_VERSION_TIVO1) {
        struct scrambleConfig config;

        prepare_scramble_regs(&config, pscramble);

        *__IDEREG32(0x9c) = 0;
        barrier();
        *__IDEREG32(0xc0) = config.wyrd[0];
        *__IDEREG32(0xc4) = config.wyrd[1];
        *__IDEREG32(0xc8) = config.wyrd[2];
        *__IDEREG32(0xcc) = config.wyrd[3];
        barrier();
        *__IDEREG32(0x9c) = 2;
    }

    unsigned long dma_base        = hwif->dma_base;
    /* write the PRD table to the controller */
    outl((unsigned long)chunk->sglist, dma_base + 4); 
    barrier();

    return 1;
}

#ifdef BCM_IDE_DEBUG
static inline void print_sg_list(struct bcm2_ide_priv *priv)
{
    int i;
    bcm2_sgent *sgent;
    int list_done;

    for (i=0; i < priv->sgmem_nfrags; i++) {
        for (sgent = priv->sgmem_frags[i], list_done=0; !list_done ;sgent++) {
            unsigned long sg_address = be32_to_cpu(sgent->address);
            unsigned long sg_count = be32_to_cpu(sgent->count);
            if (sg_count & 0x80000000) {
                sg_count &= 0x7FFFFFFF;
                list_done = 1;
            }
            printk("sg_addr = 0x%x sg_count = %d\n", sg_address, sg_count);
        }
    }
}
#endif /* BCM_IDE_DEBUG */
