/*
   Extract environment data for the makefile pre-processors.
 */

/* Copyright, compile requirements.

   (c) Keith Owens <kaos@ocs.com.au> February 2001.
   Released under GPL version 2, (you know the rest).
 */

/* Run requirements.

   The environment must contain one or more KBUILD_SRCTREE_nnn variables.
   KBUILD_SRCTREE_000 must always exist, it defines the base Linux kernel.
   Other KBUILD_SRCTREE_nnn variables are optional, gaps in the numbering
   sequence are allowed, nnn must be 000-999 with leading zeroes.

   Environment variable KBUILD_OBJTREE must be defined.
   
   KBUILD_OBJTREE can be the same as KBUILD_SRCTREE_000 if and only if
   KBUILD_SRCTREE_000 is the only source tree defined.  Otherwise all
   KBUILD_SRCTREE_nnn and KBUILD_OBJTREE names must be unique.  That is, with
   one source tree you can compile into that tree, with more than one source
   tree you must compile into a separate object tree.

   All pathnames must be absolute.

   Environment variable ARCH must be defined.
 */

#include "pp_makefile.h"

const char *srctree[MAX_SRCTREE + 1];
int high_srctree;
int common_src_obj;
const char *arch;
int arch_len;
int l_srctree[MAX_SRCTREE + 1];

/* Copy a tree name and append '/' if not already present. */
static inline void
copy_treename(const char **treename, int *savelen)
{
	int l = strlen(*treename);
	char *p = k_malloc(l + 2);
	memcpy(p, *treename, l);
	while (l-- && *(p + l) == '/') {
	};
	strcpy(p + l + 1, "/");
	*treename = p;
	*savelen = strlen(p);
}

/* Extract and validate my environment variables. */
int
pp_getenv(void)
{
	char envname[] = "KBUILD_SRCTREE_nnn";
	int i, j, high = -1;

	for (i = 0; i < MAX_SRCTREE; ++i) {
		sprintf(envname, "KBUILD_SRCTREE_%03d", i);
		if ((srctree[i] = getenv(envname))) {
			if (!*(srctree[i])) {
				srctree[i] = NULL;
				continue;
			}
			high = i;
			if (srctree[i][0] != '/') {
				fprintf(stderr,
					"Environment variable KBUILD_SRCTREE_%03d (%s) is not absolute\n",
					i, srctree[i]);
				return (1);
			}
			copy_treename(srctree + i, &l_srctree[i]);
		}
	}
	if (!srctree[0]) {
		fprintf(stderr,
			"Environment variable KBUILD_SRCTREE_000 is missing\n");
		return (1);
	}

	if (!(objtree = getenv("KBUILD_OBJTREE"))) {
		fprintf(stderr,
			"Environment variable KBUILD_OBJTREE is missing\n");
		return (1);
	}
	if (objtree[0] != '/') {
		fprintf(stderr,
			"Environment variable KBUILD_OBJTREE (%s) is not absolute\n",
			objtree);
		return (1);
	}
	copy_treename(&objtree, &l_objtree);

	for (i = 0; high && i <= high; ++i) {
		if (!srctree[i]) {
			continue;
		}
		if (!strcmp(objtree, srctree[i])) {
			fprintf(stderr,
				"Environment variable KBUILD_SRCTREE_%03d is the same as KBUILD_OBJTREE (%s)\n",
				i, srctree[i]);
			return (1);
		}
		for (j = 0; j < i; ++j) {
			if (!srctree[j]) {
				continue;
			}
			if (!strcmp(srctree[j], srctree[i])) {
				fprintf(stderr,
					"Environment variable KBUILD_SRCTREE_%03d is the same as KBUILD_SRCTREE_%03d (%s)\n",
					i, j, srctree[i]);
				return (1);
			}
		}
	}
	high_srctree = high;

	if (high == 0 && strcmp(objtree, srctree[0]) == 0) {
		common_src_obj = 1;
	}

	if (!(arch = getenv("ARCH"))) {
		fprintf(stderr, "Environment variable ARCH is missing\n");
		return (1);
	}
	arch_len = strlen(arch);

	return (0);
}

void
pp_freeenv(void)
{
	int i;
	for (i = 0; i < MAX_SRCTREE; ++i) {
		if (srctree[i]) {
			free((char *) (srctree[i]));
		}
	}
	if (objtree) {
		free((char *) objtree);
	}
}
