/*
   Standard kbuild processing for CC, CPP, AS, LD and USER targets.
 */

/* Copyright, compile requirements.

   (c) Keith Owens <kaos@ocs.com.au> March 2002.
   Released under GPL version 2, (you know the rest).

   The state machine processing was extracted from 2.4 scripts/mkdep.c.  That
   code has various authors including Linus Torvalds and Michael Elizabeth
   Chastain.
 */

/* Processing.

   To simplify the global Makefile and to make to shorter for users to do manual
   builds of kernel objects, pp_makefile5 encapsulates the standard kbuild
   processing for various targets.  The --type parameter defines the type of
   object being created, other parameters are required depending on the type.

   --target  defines the target object.  Always relative to the current directory,
             which is normally the top level object directory.
   --add     if the target name is not known until build time (e.g. using a '%.xxx'
             rule) then you need to tell pp_makefile5 to add the target to the
             database.  Otherwise it is assumed that all names are already in the
             database.
   --src     the name of the source, relative to a top level source directory.
   --srctree is the three digit source tree number.
   --flags   are all the compile flags for the object.

   --type={HOST}CC --target= [--add] --src= --srctree= --flags=

     Expands into

     rm $(target).d
     HOSTCC or no modversions
       $({HOST}CC) -Wp,-MD,$(target).d $(flags) srctree[$(srctree)]$(src) -c -o $(target) &&
	 pp_makefile5b
     not HOSTCC and modversions
       $(CPP) -Wp,-MD,$(target).d $(flags) srctree[$(srctree)]$(src) -c -o - |
	 pp_makefile5a |
	 $(CC) $(flags) -o $(target) -c -x cpp-output -

   --type=CPP --target= [--add] --src= --srctree= --flags=

     Expands into

     rm $(target).d
     $(CPP) -Wp,-MD,$(target).d $(flags) srctree[$(srctree)]$(src) -c -o $(target) &&
       pp_makefile5b

   --type={HOST}AS --target= [--add] --src= --srctree= --flags=

     Expands into

     rm $(target).d
     $({HOST}CC) -Wp,-MD,$(target).d $(flags) srctree[$(srctree)]$(src) -c -o $(target) &&
       pp_makefile5b

   --type={HOST}LD --target= [--add] --objects= --flags=

     Expands into

     rm $(target).d
     $({HOST}LD) $(flags) $(objects) -o $(target) &&
       pp_makefile5c

   --type={HOST}AR --target= [--add] --objects= --flags=

     Expands into

     rm $(target).d
     $({HOST}AR) $(flags) $(target) $(objects) &&
       pp_makefile5c

   --type=USER --target= [--add] --command= --flags

     Expands into

     rm $(target).d
     $(command) &&
       pp_makefile5d

     The flags are optional and are not used in the user command, they are used
     to detect if the flags have changed, i.e. use them to detect related config
     changes.

   pp_makefile5[abcd] are not real programs, they are all pp_makefile5 itself,
   performing different actions and using the parameters it already knows about.
   In all cases pp_makefile5 is the parent task, running other commands as
   subtasks and checking the return code from everything.  This avoids the
   problem with normal pipelines where you only see the return code from the
   last command.

   pp_makefile5a, for --type={HOST}CC.

     Read the file list.
     Read $(src), extract dependencies on CONFIG_xxx variables.
     Copy stdin to stdout (CPP output to CC input).
     Extract information from CPP data for modversions (TBD).
     After CPP has ended, close output to CC so it can finish compiling.
     When CPP has finished, read and standardize $(target).d.
     Append the config dependencies to $(target).d.
     Append the command and flags to $(target).d.
     Depending on the completion status, write the new $(target).d.

   pp_makefile5b, for --type=CPP, {HOST}AS.

     Read the file list.
     Read $(src), extract dependencies on CONFIG_xxx variables.
     When CPP has finished, read and standardize $(target).d.
     Append the config dependencies to $(target).d.
     Append the command and flags to $(target).d.
     Depending on the completion status, write the new $(target).d.

   pp_makefile5c, for --type={HOST}{LD|AR}.

     Start command running.
     While command is running, extract modversions data from the objects (TBD).
     Append the command and flags to $(target).d.
     After command has completed successfully, perform final modversion
     processing on target (TBD).
     Depending on the completion status, write the new $(target).d.

   pp_makefile5d, for --type=USER.

     Start $(command) running.
     While command is running, append the command and flags to $(target).d.
     Depending on the completion status, write the new $(target).d.

   $(target).d is the dependency file that tells make whether the target needs
   to be rebuilt on the next kbuild run.  It is written by CPP (using option
   -Wp,-MD,$(target).d).  After the step runs successfully, pp_makefile5 reads
   the .d file, copies the data into the pre-processor database and removes the
   .d file.  gcc cannot handle -Wx parameters that contain embedded commas and I
   have not found any quoting method to work around this.  So ',' in the .d
   filename is changed to ':'.

   CPP writes the dependency list of files that were actually included, this is
   accurate because it includes .c as well as .h files and because it includes
   exactly the files that were read, no more and no less.  However the .d file
   created by CPP has some deficiencies for kbuild.  CPP assumes that the source
   is in the current directory so the first line is always $(notdir $(target)).
   The included files have absolute path names, kbuild wants relative names for
   files in the object directory and variable names for files in source
   directories.  pp_makefile5 reads $(target).d after CPP has completed
   successfully and, in memory :-

     Deletes the target information.
     Changes absolute path names for object and source tree files to relative
     path names, also recording the tree index.
     Changes .tmp_include/src_nnn/asm/ to include/asm-$(ARCH)/ and index.
     Changes other .tmp_include/src_nnn/ to include/ and index.
     Discards include/linux/autoconf.h.
     Leaves relative path names alone, they are valid for the object tree.
     Discards other absolute paths, they are outside kbuild, e.g. stdarg.h.

   The command and flags are written to the database, so that make can detect if
   anything has changed and force a rebuild.  For example

     ifneq ($(KCN_$(target)),$(KCO_$(target)))
       $(target): FORCE_BUILD_COMMAND
     endif

   KCN_$(target) (new commands) is set for each target in the global makefile.
   KCO_$(target) (old commands) is set for each target by pp_makefile5.
   The components of KC[NO] depend on the --type.

     --type={HOST}CC		$({HOST}CC) $(flags)
     --type=CPP			$(CPP) $(flags)
     --type={HOST}AS		$({HOST}AS) $(flags)
     --type={HOST}LD		$({HOST}LD) $(objects) $(flags)
     --type={HOST}AR		$({HOST}AR) $(objects) $(flags)
     --type=USER		$(command) $(flags)

   If there any dependency entries in $(target).d, a check is made for the
   presence of config data for each dependency.  If the dependency has no config
   data or is not marked as uptodate then the dependency is read and references
   to CONFIG variables are saved.
 */

/* Parameters.

   --type, --target, --add, --src, --srctree, --flags, --command.
   See above for valid combinations.

   --verbose.  Display what pp_makefile5 is doing.

   --debug.  Activate some features that make it easier to debug pp_makefile5.

 */

#include "pp_makefile.h"
#include "pp_db.h"

#include <ctype.h>
#include <errno.h>
#include PP_GETOPT_H
#include <limits.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/wait.h>

const char *program = "pp_makefile5";

int errors;
int norun;
int verbose;
int signalled;

static int debug;
static int add;

static enum {
	TYPE_CC,
	TYPE_HOSTCC,
	TYPE_CPP,
	TYPE_AS,
	TYPE_HOSTAS,
	TYPE_LD,
	TYPE_HOSTLD,
	TYPE_AR,
	TYPE_HOSTAR,
	TYPE_USER,
} type;

#define is_compile_type()	(type == TYPE_CC || type == TYPE_HOSTCC || type == TYPE_CPP || type == TYPE_AS || type == TYPE_HOSTAS)
#define is_link_type()		(type == TYPE_LD || type == TYPE_HOSTLD || type == TYPE_AR || type == TYPE_HOSTAR)
#define is_user_type()		(type == TYPE_USER)

#define OPTION_TYPE	(1 << 0)
#define OPTION_TARGET	(1 << 1)
#define OPTION_SRC	(1 << 2)
#define OPTION_SRCTREE	(1 << 3)
#define OPTION_FLAGS	(1 << 4)
#define OPTION_OBJECTS	(1 << 5)
#define OPTION_COMMAND	(1 << 6)
static int option;

static const char *option_type;
static const char *option_target;
static const char *option_src;
static const char *option_srctree;
static const char *option_flags;
static const char *option_objects;
static const char *option_command;

static char *option_target2;

static int l_target;
static int l_src;
static int l_flags;
static int l_objects;
static int l_command;

static const char *this_srctree;
static int option_srctree_n;

/* Record the use of a config entry */

static void
use_config(PPDB * db, DB_TOKEN token, DB_REC_SOURCE * s, const char *start,
	   const char *next)
{
	DB_REC_SOURCE *s_config;
	DB_REC_CCD_1 add, *c1;
	DB_TOKEN token_config;
	char name[PATH_MAX + 1], *p;
	strcpy(name, CONFIG_SETTINGS);
	for (p = name + strlen(name);
	     start < next && p < name + sizeof (name) - 3; ++p, ++start) {
		*p = tolower(*start);
		if (*p == '_') {
			*p = '/';
		}
	}
	strcpy(p, ".h");
	token_config = lookup_db_byname_highest(db, name, DB_RECTYPE_SOURCE);
	if (DB_IS_NULL_TOKEN(token_config)) {
		findnew_db_target(db, name);
		token_config =
		    lookup_db_byname(db, name, DB_RECTYPE_SOURCE,
				     objtree_treename_index);
		assert(!DB_IS_NULL_TOKEN(token_config));
	}
	s_config = read_db_source(db, token_config);
	add.entry = token_config;
	add_db_ccd_list(db, s->config, &add, &c1);
	c1->index = s_config->index;
	c1->mtime = s_config->mtime;
}

/* State machine extracted from scripts/mkdep.c.  Unlike mkdep, this program
 * only looks for CONFIG_ entries, #include is handled by cpp and is output in
 * the .d dependency file.
 */

/*
 * Macros for stunningly fast map-based character access.
 * __buf is a register which holds the current word of the input.
 * Thus, there is one memory access per sizeof(unsigned long) characters.
 */

#if defined(__alpha__) || defined(__i386__) || defined(__ia64__) || defined(__MIPSEL__)	\
    || defined(__arm__)
#define LE_MACHINE
#endif

#ifdef LE_MACHINE
#define next_byte(x) (x >>= 8)
#define current ((unsigned char) __buf)
#else
#define next_byte(x) (x <<= 8)
#define current (__buf >> 8*(sizeof(unsigned long)-1))
#endif

#define GETNEXT { \
	next_byte(__buf); \
	if ((unsigned long) next % sizeof(unsigned long) == 0) { \
		if (next >= end) \
			break; \
		__buf = * (unsigned long *) next; \
	} \
	next++; \
}

/*
 * State machine macros.
 */
#define CASE(c,label) if (current == c) goto label
#define NOTCASE(c,label) if (current != c) goto label

/*
 * Yet another state machine speedup.
 */
#define MAX2(a,b) ((a)>(b)?(a):(b))
#define MIN2(a,b) ((a)<(b)?(a):(b))
#define MAX5(a,b,c,d,e) (MAX2(a,MAX2(b,MAX2(c,MAX2(d,e)))))
#define MIN5(a,b,c,d,e) (MIN2(a,MIN2(b,MIN2(c,MIN2(d,e)))))

/*
 * The state machine looks for (approximately) these Perl regular expressions:
 *
 *    m|\/\*.*?\*\/|
 *    m|\/\/.*|
 *    m|'.*?'|
 *    m|".*?"|
 *    m|#\s*(?define|undef)\s*CONFIG_(\w*)|
 *    m|(?!\w)CONFIG_|
 *
 * About 98% of the CPU time is spent here, and most of that is in
 * the 'start' paragraph.  Because the current characters are
 * in a register, the start loop usually eats 4 or 8 characters
 * per memory read.  The MAX5 and MIN5 tests dispose of most
 * input characters with 1 or 2 comparisons.
 */
static void
state_machine(PPDB * db, DB_TOKEN token, DB_REC_SOURCE * s, const char *map,
	      const char *end)
{
	const char *next = map;
	const char *map_dot;
	unsigned long __buf = 0;

	for (;;) {
	      start:
	      GETNEXT __start:
		if (current > MAX5('/', '\'', '"', '#', 'C'))
			goto start;
		if (current < MIN5('/', '\'', '"', '#', 'C'))
			goto start;
		CASE('/', slash);
		CASE('\'', squote);
		CASE('"', dquote);
		CASE('#', pound);
		CASE('C', cee);
		goto start;

/* // */
	      slash_slash:
		GETNEXT CASE('\n', start);
		NOTCASE('\\', slash_slash);
		GETNEXT goto slash_slash;

/* / */
	      slash:
		GETNEXT CASE('/', slash_slash);
		NOTCASE('*', __start);
	      slash_star_dot_star:
	      GETNEXT __slash_star_dot_star:
		NOTCASE('*', slash_star_dot_star);
		GETNEXT NOTCASE('/', __slash_star_dot_star);
		goto start;

/* '.*?' */
	      squote:
		GETNEXT CASE('\'', start);
		NOTCASE('\\', squote);
		GETNEXT goto squote;

/* ".*?" */
	      dquote:
		GETNEXT CASE('"', start);
		NOTCASE('\\', dquote);
		GETNEXT goto dquote;

/* #\s* */
	      pound:
		GETNEXT CASE(' ', pound);
		CASE('\t', pound);
		CASE('d', pound_d);
		CASE('u', pound_u);
		goto __start;

/* #\s*d */
	      pound_d:
		GETNEXT NOTCASE('e', __start);
		GETNEXT NOTCASE('f', __start);
		GETNEXT NOTCASE('i', __start);
		GETNEXT NOTCASE('n', __start);
		GETNEXT NOTCASE('e', __start);
		goto pound_define_undef;

/* #\s*u */
	      pound_u:
		GETNEXT NOTCASE('n', __start);
		GETNEXT NOTCASE('d', __start);
		GETNEXT NOTCASE('e', __start);
		GETNEXT NOTCASE('f', __start);
		goto pound_define_undef;

/*
 * #\s*(define|undef)\s*CONFIG_(\w*)
 *
 * this does not define the word, because it could be inside another
 * conditional (#if 0).  But I do parse the word so that this instance
 * does not count as a use.  -- mec
 */
	      pound_define_undef:
		GETNEXT CASE(' ', pound_define_undef);
		CASE('\t', pound_define_undef);

		NOTCASE('C', __start);
		GETNEXT NOTCASE('O', __start);
		GETNEXT NOTCASE('N', __start);
		GETNEXT NOTCASE('F', __start);
		GETNEXT NOTCASE('I', __start);
		GETNEXT NOTCASE('G', __start);
		GETNEXT NOTCASE('_', __start);

		map_dot = next;
	      pound_define_undef_CONFIG_word:
		GETNEXT if (isalnum(current) || current == '_')
			goto pound_define_undef_CONFIG_word;
		goto __start;

/* \<CONFIG_(\w*) */
	      cee:
		if (next >= map + 2 && (isalnum(next[-2]) || next[-2] == '_'))
			goto start;
		GETNEXT NOTCASE('O', __start);
		GETNEXT NOTCASE('N', __start);
		GETNEXT NOTCASE('F', __start);
		GETNEXT NOTCASE('I', __start);
		GETNEXT NOTCASE('G', __start);
		GETNEXT NOTCASE('_', __start);

		map_dot = next;
	      cee_CONFIG_word:
		GETNEXT if (isalnum(current) || current == '_')
			goto cee_CONFIG_word;
		use_config(db, token, s, map_dot, next - 1);
		goto __start;
	}
}

/* Read a source file and run the state machine to extract config usage. */

static int
extract_config(PPDB * db, DB_TOKEN token, DB_REC_SOURCE * s)
{
	int mapsize, fd, ret = 1, pagesizem1 = getpagesize() - 1;
	struct stat st;
	char *map, name[PATH_MAX + 1];

	snprintf(name, sizeof (name), "%s%s", srctree[s->index],
		 read_db_text(db, s->relname) + 1);
	fd = open(name, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "%s: open(%s) failed: %m\n", program, name);
		goto out1;
	}
	if (fstat(fd, &st)) {
		fprintf(stderr, "%s: fstat(%s) failed: %m\n", program, name);
		goto out2;
	}
	if (st.st_size == 0) {
		fprintf(stderr, "%s: %s is empty: %m\n", program, name);
		goto out2;
	}

	mapsize = st.st_size;
	mapsize = (mapsize + pagesizem1) & ~pagesizem1;
	map = mmap(NULL, mapsize, PROT_READ, MAP_SHARED, fd, 0);
	if ((long) map == -1) {
		fprintf(stderr, "%s: mmap(%s) failed: %m\n", program, name);
		goto out2;
	}
	if ((unsigned long) map % sizeof (unsigned long) != 0) {
		fprintf(stderr, "%s: mmap(%s) not aligned: %m\n", program,
			name);
		goto out3;
	}

	state_machine(db, token, s, map, map + st.st_size);

	ret = 0;

      out3:
	munmap(map, mapsize);
      out2:
	close(fd);
      out1:
	return (ret);
}

/* Modify text with correct quoting for make to read it back */

static void
clean_make(char **text)
{
	int l = 2 * strlen(*text) + 1;
	char *text2 = k_malloc(l), *p = text2, *text_orig = *text;
	while (**text) {
		if (**text == '#') {
			*p++ = '\\';
		} else if (**text == '$' && memcmp(*text, "$(KBUILD_", 9)) {
			*p++ = '$';
		}
		*p++ = **text;
		(*text)++;
	}
	*p++ = '\0';
	assert(strlen(text2) < l);
	free(text_orig);
	*text = text2;
}

/*============================================================================*/

static void
usage(void)
{
	fprintf(stderr,
		"usage: %s --type={HOST}{CC|AS} --target= [--add] --src= --srctree= --flags=\n",
		program);
	fprintf(stderr,
		"       %s --type=CPP --target= [--add] --src= --srctree= --flags=\n",
		program);
	fprintf(stderr,
		"       %s --type={HOST}{LD|AR} --target= [--add] --objects= --flags=\n",
		program);
	fprintf(stderr,
		"       %s --type=USER --target= [--add] --command= [--flags=]\n",
		program);
	fprintf(stderr,
		"       All of the above also accept --norun, --verbose and --debug\n");
	exit(1);
}

static int
parse(int argc, char **argv)
{
	int o;
	static struct option long_opts[] = {
		{"type", 1, 0, 't'},
		{"target", 1, 0, 'g'},
		{"add", 0, 0, 'a'},
		{"src", 1, 0, 's'},
		{"srctree", 1, 0, 'N'},
		{"flags", 1, 0, 'f'},
		{"objects", 1, 0, 'o'},
		{"command", 1, 0, 'c'},
		{"norun", 0, 0, 'n'},
		{"verbose", 0, 0, 'v'},
		{"debug", 0, 0, 'd'},
		{0, 0, 0, 0}
	};

	/* Parse the command line.  */
	while ((o = getopt_long(argc, argv, "t:ga:s:N:f:o:c:nvd",
				&long_opts[0], NULL)) != EOF) {
		switch (o) {
		case 't':	/* --type */
			option |= OPTION_TYPE;
			option_type = optarg;
			if (strcmp(optarg, "CC") == 0) {
				type = TYPE_CC;
			} else if (strcmp(optarg, "HOSTCC") == 0) {
				type = TYPE_HOSTCC;
			} else if (strcmp(optarg, "CPP") == 0) {
				type = TYPE_CPP;
			} else if (strcmp(optarg, "AS") == 0) {
				type = TYPE_AS;
			} else if (strcmp(optarg, "HOSTAS") == 0) {
				type = TYPE_HOSTAS;
			} else if (strcmp(optarg, "LD") == 0) {
				type = TYPE_LD;
			} else if (strcmp(optarg, "HOSTLD") == 0) {
				type = TYPE_HOSTLD;
			} else if (strcmp(optarg, "AR") == 0) {
				type = TYPE_AR;
			} else if (strcmp(optarg, "HOSTAR") == 0) {
				type = TYPE_HOSTAR;
			} else if (strcmp(optarg, "USER") == 0) {
				type = TYPE_USER;
			} else {
				fprintf(stderr,
					"%s: --type must be one of {HOST}CC, CPP, {HOST}AS, {HOST}LD, {HOST}AR or USER.  %s is invalid\n",
					program, optarg);
				++errors;
			}
			break;
		case 'g':	/* --target */
			option |= OPTION_TARGET;
			option_target = optarg;
			break;
		case 'a':	/* --add */
			add = 1;
			break;
		case 's':	/* --src */
			option |= OPTION_SRC;
			option_src = optarg;
			break;
		case 'N':	/* --srctree */
			option |= OPTION_SRCTREE;
			option_srctree = optarg;
			break;
		case 'f':	/* --flags */
			option |= OPTION_FLAGS;
			option_flags = optarg;
			break;
		case 'o':	/* --objects */
			option |= OPTION_OBJECTS;
			option_objects = optarg;
			break;
		case 'c':	/* --command */
			option |= OPTION_COMMAND;
			option_command = optarg;
			break;
		case 'n':	/* --norun */
			norun = 1;
			break;
		case 'v':	/* --verbose */
			++verbose;
			break;
		case 'd':	/* --debug */
			debug = 1;
			break;
		default:
			usage();
			break;
		}
	}

	if (optind != argc) {
		usage();
	}

	if (is_compile_type()) {
		if (option !=
		    (OPTION_TYPE | OPTION_TARGET | OPTION_SRC | OPTION_SRCTREE |
		     OPTION_FLAGS)) {
			usage();
		}
	} else if (is_link_type()) {
		if (option !=
		    (OPTION_TYPE | OPTION_TARGET | OPTION_OBJECTS |
		     OPTION_FLAGS)) {
			usage();
		}
	} else if (is_user_type()) {
		if ((option | OPTION_FLAGS) !=
		    (OPTION_TYPE | OPTION_TARGET | OPTION_COMMAND |
		     OPTION_FLAGS)) {
			usage();
		}
	} else {
		usage();
	}

	if (!*option_target || *option_target == '/') {
		fprintf(stderr, "%s: target must be a relative filename\n",
			program);
		++errors;
	}

	if (option_src && (!*option_src || *option_src == '/')) {
		fprintf(stderr, "%s: src must be a relative filename\n",
			program);
		++errors;
	}

	if (option_srctree &&
	    !(isdigit(option_srctree[0]) &&
	      isdigit(option_srctree[1]) &&
	      isdigit(option_srctree[2]) && option_srctree[3] == '\0')) {
		fprintf(stderr, "%s: srctree must be three digits\n", program);
		++errors;
	}

	if (option_objects
	    && !*(option_objects + strspn(option_objects, " \t\n"))) {
		fprintf(stderr, "%s: objects must be non-blank\n", program);
		++errors;
	}

	if (option_command
	    && !*(option_command + strspn(option_command, " \t\n"))) {
		fprintf(stderr, "%s: command must be non-blank\n", program);
		++errors;
	}

	if (option_objects && !*option_objects) {
		fprintf(stderr, "%s: objects must be supplied\n", program);
		++errors;
	}

	l_target = strlen(option_target);
	option_target2 = k_malloc(l_target + 2);
	option_target2[0] = '/';
	memcpy(option_target2 + 1, option_target, l_target + 1);
	l_src = option_src ? strlen(option_src) : 0;
	l_flags = option_flags ? strlen(option_flags) : 0;
	l_objects = option_objects ? strlen(option_objects) : 0;
	l_command = option_command ? strlen(option_command) : 0;

	if (option_srctree) {
		option_srctree_n = strtoul(option_srctree, NULL, 10);
		this_srctree = srctree[option_srctree_n];
	}

	if (!errors && !verbose) {
		char *quiet;
		if ((quiet = getenv("KBUILD_QUIET")) && strcmp(quiet, "@") == 0) {
			fprintf(stdout, "  %s %s\n", option_type,
				option_target);
		}
	}

	return (errors);
}

static const char *
readenv(const char *name)
{
	const char *env = getenv(name);
	if (!env) {
		fprintf(stderr, "%s: No environment value for %s\n", program,
			name);
		++errors;
	}
	return (env);
}

/* Fork a command. */

static pid_t
fork_command(int fd0, int fd1, int close_child, int close_parent,
	     const char *command, ...)
{
	int lt = strlen(command), ls, size = lt + 4096;
	char *text = k_malloc(size), *s;
	pid_t pid;
	va_list ap;

	do {
		lt = strlen(command);
		memcpy(text, command, lt);
		text[lt++] = ' ';
		va_start(ap, command);
		while ((s = va_arg(ap, char *))) {
			ls = strlen(s);
			if (lt + ls >= size - 1) {
				text = k_realloc(text, size += 4096);
				break;
			}
			memcpy(text + lt, s, ls);
			lt += ls;
		}
	} while (s);
	text[lt] = '\0';
	va_end(ap);

	if ((pid = fork()) == 0) {
		/* child */
		if (debug) {
			sleep(10);	/* time for gdb connect */
		}
		if (fd0 >= 0 && fd0 != 0) {
			close(0);
			if (dup(fd0) < 0) {
				fprintf(stderr, "%s: dup failed: %m\n",
					program);
				exit(1);
			}
			VERBOSE(2, "child: dup %d to 0\n", fd0);
		}
		if (fd1 >= 0 && fd1 != 1) {
			close(1);
			if (dup(fd1) < 0) {
				fprintf(stderr, "%s: dup failed: %m\n",
					program);
				exit(1);
			}
			VERBOSE(2, "child: dup %d to 1\n", fd1);
		}
		if (close_child >= 0) {
			close(close_child);
		}
		VERBOSE(2, "close_child: %d\n", close_child);
		execl("/bin/sh", "/bin/sh", "-c", text, NULL);
		fprintf(stderr, "%s: execl(/bin/sh -c %s) failed %m\n", program,
			text);
		exit(1);
	} else if (pid < 0) {
		fprintf(stderr, "%s: fork failed %m\n", program);
		++errors;
	} else if (debug) {
		fprintf(stderr, "%s: fork_command child pid %d\n", program,
			pid);
	}
	if (close_parent >= 0) {
		close(close_parent);
	}
	VERBOSE(2, "close_parent: %d\n", close_parent);
	VERBOSE(1, "  %s", text);	/* Newline written by caller */
	free(text);
	return (pid);
}

/* Check that the config entries for a file are up to date. */

static void
check_config(PPDB * db, DB_TOKEN token, DB_REC_SOURCE * s)
{
	DB_REC_CCD_LIST *c;
	if (DB_IS_NULL_TOKEN(s->config)) {
		new_db_ccd_list(db, &(s->config));
	}
	c = read_db_ccd_list(db, s->config);
	VERBOSE(3, "    check_config %s %d/%ld %d/%ld\n",
		read_db_text(db, s->relname), s->index, s->mtime,
		c->parent_index, c->parent_mtime);
	if (c->parent_mtime == s->mtime && c->parent_index == s->index) {
		return;
	}
	/* extract_config() can expand the ccd_list, c is unreliable */
	c = NULL;
	if (extract_config(db, token, s)) {
		VERBOSE(2, "  extract_config %s failed\n",
			read_db_text(db, s->relname) + 1);
		++errors;
	} else {
		c = read_db_ccd_list(db, s->config);	/* may have moved */
		c->parent_index = s->index;
		c->parent_mtime = s->mtime;
	}
	VERBOSE(3, "    check_config %d/%ld %d/%ld\n", s->index, s->mtime,
		c->parent_index, c->parent_mtime);
}

/* Dependency work is only executed in the dependency subtask, after the main
 * task has said whether the command worked or not.
 */

/* Standardize a single name.  Ugly code here, avert your eyes.
 *
 * If the name can be resolved to the object or one of the source trees then
 * strip the variable start of the name to give a relative name.  include/asm/
 * is converted to include/asm-$(ARCH)/ to get the real name instead of the
 * symlink.  index is set to the number of the object or source tree, -1 if
 * the name could not be resolved to object or source.
 */

static char *
standardize_name(PPDB * db, const char *name, int *index, int must_exist)
{
	char *name2 = k_strdup(name);
	int i, lname2 = strlen(name2);

	*index = -1;

	/* Normalize /../ and remove multiple /  */
	if (strstr(name2, "/../") || strstr(name2, "//")) {
		char *p;
		for (p = name2; (p = strstr(p, "//")); strcpy(p + 1, p + 2)) {
		};
		while ((p = strstr(name2, "/../"))) {
			if (p == name2) {
				strcpy(name2 + 1, name2 + 4);
			} else {
				char *p1;
				for (p1 = p - 1; *p1 != '/' && p1 >= name2;
				     --p1) {
				};
				strcpy(p1 + 1, p + 4);
			}
		}
		for (p = name2; (p = strstr(p, "//")); strcpy(p + 1, p + 2)) {
		};
		lname2 = strlen(name2);
	}

	/* Try to map absolute names to $(KBUILD_OBJTREE) or $(KBUILD_SRCTREE_nnn) */
	if (*name2 == '/') {
		for (i = objtree_treename_index; i >= 0; --i) {
			if (l_srctree[i] && lname2 > l_srctree[i] &&
			    memcmp(name2, srctree[i], l_srctree[i]) == 0) {
				char *name3 = k_strdup(name2 + l_srctree[i]);	/* $(KBUILD_SRCTREE_nnn)foo -> foo */
				*index = i;
				free(name2);
				name2 = name3;
				lname2 = strlen(name2);
				break;
			}
			/* Jump from objtree to highest srctree, save time */
			if (i > high_srctree) {
				i = high_srctree + 1;
			}
		}
	}

	/* Strip relative paths starting with '.tmp_include/src_nnn/' and set index
	 * to nnn.
	 */
	if (*name2 != '/') {
		static char tmp_include[] = ".tmp_include/src_nnn/";
		const int linclude = sizeof (tmp_include) - 1;
		if (lname2 > linclude &&
		    memcmp(name2, tmp_include, linclude - 4) == 0 &&
		    *(name2 + linclude - 1) == '/') {
			char *p, *name3, *nnn;
			name3 = k_strdup(name2 + linclude);	/* .tmp_include/src_nnn/foo -> foo */
			nnn = name2 + linclude - 4;
			nnn[3] = '\0';
			*index = strtoul(nnn, &p, 10);
			if (*p) {
				fprintf(stderr,
					"%s: Invalid index '%s' in %s\n",
					program, nnn, name);
				++errors;
				*index = -1;
			}
			free(name2);
			name2 = name3;
			lname2 = strlen(name2);
		}
	}

	/* Convert relative paths with 'include/asm/' to 'include/asm-$(ARCH)/'. */
	if (*name2 != '/') {
		static char include_asm[] = "include/asm/";
		const int linclude = sizeof (include_asm) - 1;
		char *p, *p3;
		while ((p = strstr(name2, include_asm))) {
			char *name3 = k_malloc(lname2 + arch_len + 2);
			memcpy(name3, name2, (p - name2) + linclude - 1);	/* ...include/asm */
			p3 = name3 + (p - name2) + linclude - 1;
			*p3 = '-';	/* ...include/asm- */
			memcpy(p3 + 1, arch, arch_len);	/* ...include/asm-$(ARCH) */
			strcpy(p3 + 1 + arch_len, p + linclude - 1);	/* ...include/asm-$(ARCH)/... */
			free(name2);
			name2 = name3;
			lname2 = strlen(name2);
		}
	}

	/* Verify that the name is known, ignoring unconverted absolute names.  Handle
	 * common source and object trees, the file was assigned to the object tree
	 * earlier but may really be source, we could not be sure until here.
	 */
	if (*name2 != '/') {
		char name3[lname2 + 2];
		DB_TOKEN token;
		DB_REC_SOURCE *s;
		name3[0] = '/';
		memcpy(name3 + 1, name2, lname2 + 1);
		token = lookup_db_byname_highest(db, name3, DB_RECTYPE_SOURCE);
		if (DB_IS_NULL_TOKEN(token)) {
			s = NULL;
			*index = -1;
		} else if (*index == objtree_treename_index || *index == -1) {
			s = read_db_source(db, token);
			*index = s->index;
		}
		if (*index == -1 && must_exist) {
			fprintf(stderr,
				"%s: standardize_name: %s (%s) not in filelist, should never happen\n",
				program, name, name3);
			++errors;
		}
	}

	/* The caller must free the result */
	return (name2);
}

/* Standardize absolute filenames in the flags and command, e.g. include paths.
 * An absolute name is '/' at start of text, '<whitespace>/', '=/', '-./' at
 * start of text or '<whitespace>-./'.
 */

static char *
standardize_names(PPDB * db, const char *original_text)
{
	static char kbuild_srctree_nnn[] = "$(KBUILD_SRCTREE_nnn)";
	char *text = k_strdup(original_text), *p =
	    text, *new_text, *p1, *abs, c;
	int s1, s2a, s2b, s3, index;
	while ((abs = strchr(p, '/'))) {
		if (abs == text || isspace(*(abs - 1)) || *(abs - 1) == '=' ||
		    (abs >= text + 2 && *(abs - 2) == '-'
		     && (abs == text + 2 || isspace(*(abs - 3))))) {
			s1 = abs - text;
			for (p1 = abs;
			     *p1 && !isspace(*p1) && *p1 != ')' && *p1 != ';';
			     ++p1) {
			};
			s3 = strlen(p1);
			c = *p1;
			*p1 = '\0';
			abs = standardize_name(db, abs, &index, 0);
			*p1 = c;
			s2a = (index < 0 || index == objtree_treename_index)
			    ? 0 : sizeof (kbuild_srctree_nnn) - 1;
			s2b = strlen(abs);
			new_text = k_malloc(s1 + s2a + s2b + s3 + 1);
			memcpy(new_text, text, s1);
			if (s2a) {
				sprintf(kbuild_srctree_nnn +
					sizeof (kbuild_srctree_nnn) - 1 - 4,
					"%03d)", index);
				memcpy(new_text + s1, kbuild_srctree_nnn, s2a);
			}
			memcpy(new_text + s1 + s2a, abs, s2b);
			memcpy(new_text + s1 + s2a + s2b, p1, s3 + 1);
			free(abs);
			free(text);
			text = new_text;
			p = text + s1 + s2a + s2b;
		} else {
			++p;	/* Not an absolute name */
		}
	}
	return (text);
}

/* Extract and standardize dependency filenames from a single line.
 * Simplistic tokenising, extract names delimited by white space, '\' or ':'.
 * No kernel filenames should ever contain these characters.
 */

struct dep {
	char *name;
	int index;
};

static void
extract_dependencies(PPDB * db, char *line, struct dep **dep_list,
		     int *dep_size, int *dep_used)
{
	char *name, *new_name;
	int index;

	while (1) {
		line += strspn(line, " \t\n\\");
		if (!*line) {
			return;
		}

		name = line;
		line += strcspn(line, ": \t\n\\");
		if (*line == ':') {
			/* 'target(s):', discard any saved names */
			while (*dep_used > 0) {
				--*dep_used;
				free((*dep_list)[*dep_used].name);
				(*dep_list)[*dep_used].name = NULL;
			}
			++line;
			continue;
		}

		/* Probably a dependency name, standardize it */
		*(line++) = '\0';	/* terminate name */
		new_name = standardize_name(db, name, &index, 1);
		if (index < 0
		    || strcmp(new_name, "include/linux/autoconf.h") == 0) {
			/* Ignore unconverted paths and include/linux/autoconf.h */
			;
		} else {
			if (*dep_used + 1 >= *dep_size) {
				*dep_list =
				    k_realloc(*dep_list,
					      ((*dep_size +=
						100)) * sizeof (**dep_list));
			}
			(*dep_list)[*dep_used].name =
			    k_malloc(strlen(new_name) + 2);
			(*dep_list)[*dep_used].name[0] = '/';
			strcpy((*dep_list)[*dep_used].name + 1, new_name);
			free(new_name);
			(*dep_list)[*dep_used].index = index;
			(*dep_list)[++*dep_used].name = NULL;
		}
	}
}

/* Read the dependencies written by cpp */

static int
read_depend(PPDB * db, const char *depend, struct dep **dep)
{
	FILE *fdep;
	char *line;
	int line_size, eof = 0, dep_size = 0, dep_used = 0;
	*dep = NULL;
	if (!(fdep = fopen(depend + 1, "r"))) {
		fprintf(stderr, "%s: Cannot fopen(%s) for reading: %m\n",
			program, depend + 1);
		return (1);
	}
	line = k_malloc((line_size = 4096));
	while (!(eof = fgets_long_line(&line, &line_size, fdep))) {
		extract_dependencies(db, line, dep, &dep_size, &dep_used);
	}
	free(line);
	fclose(fdep);
	if (eof < 0) {
		fprintf(stderr, "%s: Error reading from %s: %m\n", program,
			depend + 1);
		return (1);
	}
	unlink(depend + 1);
	if (unlink(depend + 1) && errno != ENOENT) {
		fprintf(stderr, "%s: Cannot unlink %s: %m\n", program,
			depend + 1);
		return (1);
	}
	return (0);
}

/* Some dependencies are generated during the build phase so their timestamp is
 * not known until phase5.  Update the database with the timestamps for command
 * generated dependencies.  The filenames must already be in the database as
 * targets, otherwise we have an incomplete dependency graph.
 */

static void
new_dependency_timestamp(PPDB * db, DB_REC_SOURCE * s, const char *relname)
{
	DB_TOKEN token;
	DB_REC_TARGET *t;
	struct stat statbuf;
	token = lookup_db_target(db, relname);
	if (DB_IS_NULL_TOKEN(token)) {
		fprintf(stderr,
			"%s: User error, %s is used as a dependency but is not specified in the makefiles, incomplete dependency graph!\n",
			program, relname + 1);
		++errors;
		return;
	}
	t = read_db_target(db, token);
	if (stat(relname + 1, &statbuf)) {
		fprintf(stderr, "%s: stat (%s) failed: %m\n", program,
			relname + 1);
		++errors;
		return;
	}
	/* Micro optimization, only dirty the page if we have to, reduce writeback */
	if (t->mtime != statbuf.st_mtime) {
		t->mtime = statbuf.st_mtime;
	}
	if (s->mtime != statbuf.st_mtime) {
		s->mtime = statbuf.st_mtime;
	}
}

/* Standardize and save the complete dependency data */

static int
write_depend(PPDB * db, const char *command, const char *depend)
{
	DB_TOKEN token;
	DB_REC_SOURCE *s;
	DB_REC_TARGET *t;
	DB_REC_CCD_1 add, *c1;
	struct dep *dep_list = NULL, *dep;
	struct stat statbuf;
	char *stand_command = standardize_names(db, command);
	char *stand_flags = standardize_names(db, option_flags ? option_flags : "");
	int ret = 0;
	int l =
	    strlen(stand_command) + strlen(stand_flags) + 200 +
	    (is_link_type()? l_objects : 0);
	char buf[l];

	/* Record new timestamp for target */
	if (stat(option_target2 + 1, &statbuf)) {
		fprintf(stderr, "%s: stat (%s) failed: %m\n", program,
			option_target2 + 1);
		statbuf.st_mtime = 0;
	}
	token =
	    lookup_db_byname(db, option_target2, DB_RECTYPE_SOURCE,
			     objtree_treename_index);
	assert(!DB_IS_NULL_TOKEN(token));
	s = read_db_source(db, token);
	s->mtime = statbuf.st_mtime;
	token = lookup_db_target(db, option_target2);
	assert(!DB_IS_NULL_TOKEN(token));
	t = read_db_target(db, token);
	t->mtime = statbuf.st_mtime;
	if (!t->mtime) {
		return (1);
	}

	/* Read and extract the dependency list and copy into the database */
	if (!t->flags.nodepend && !t->flags.commands
	    && DB_IS_NULL_TOKEN(t->objlink)) {
		if (read_depend(db, depend, &dep_list)) {
			return (1);
		}
		if (DB_IS_NULL_TOKEN(t->dependency)) {
			new_db_ccd_list(db, &(t->dependency));
			t = read_db_target(db, token);	/* may have moved */
		} else {
			empty_db_ccd_list(db, t->dependency);
		}
		memset(&add, 0, sizeof (add));
	}
	for (dep = dep_list; dep && dep->name; ++dep) {
		DB_TOKEN dep_token;
		int commands = 0;
		dep_token = lookup_db_target(db, dep->name);
		if (!DB_IS_NULL_TOKEN(dep_token)) {
			DB_REC_TARGET *dep_t = read_db_target(db, dep_token);
			commands = dep_t->flags.commands
			    || dep_t->flags.nodepend;
		}
		dep_token =
		    lookup_db_byname(db, dep->name, DB_RECTYPE_SOURCE,
				     dep->index);
		assert(!DB_IS_NULL_TOKEN(dep_token));
		s = read_db_source(db, dep_token);
		if (commands) {
			new_dependency_timestamp(db, s, dep->name);
		}
		add.entry = dep_token;
		add_db_ccd_list(db, t->dependency, &add, &c1);
		c1->index = s->index;
		c1->mtime = s->mtime;
		check_config(db, dep_token, s);
		t = read_db_target(db, token);	/* may have moved */
		free(dep->name);
	}
	if (dep_list) {
		DB_REC_CCD_LIST *c = read_db_ccd_list(db, t->dependency);
		c->parent_mtime = t->mtime;
		free(dep_list);
	}

	/* Set the command in the database */
	clean_make(&stand_command);
	clean_make(&stand_flags);
	if (is_compile_type() || is_user_type()) {
		snprintf(buf, sizeof (buf), "%s %s", stand_command,
			 stand_flags);
	} else if (is_link_type()) {
		snprintf(buf, sizeof (buf), "%s %s %s", stand_command,
			 option_objects, stand_flags);
	} else {
		snprintf(buf, sizeof (buf), "unknown type:%d %s %s", type,
			 stand_command, stand_flags);
		ret = 1;
	}
	free(stand_command);
	free(stand_flags);
	if (ret == 0) {
		if (DB_IS_NULL_TOKEN(t->command)) {
			new_db_text(db, buf, &(t->command));
		} else {
			update_db_text(db, buf, t->command);
		}
		t = read_db_target(db, token);	/* may have moved */
		t->flags.uptodate = 1;
	}
	return (ret);
}

/* Starts off running in parallel with compile/link and command subtasks then
 * blocks until the main task says what the final command status is.
 */

static int
do_dependencies(int fd, const char *command, const char *depend)
{
	PPDB *db_main = NULL;
	fd_set rfds;
	struct timeval tv;
	int close_it, l, ret = 0;
	char ok;
	VERBOSE(1, "do_dependencies\n");
	db_main = open_db_main(1);
	if (!db_main) {
		VERBOSE(2, "  do_dependencies open_db_main failed\n");
		return (1);
	}
	if (add) {
		findnew_db_target(db_main, option_target2);
	}
	/* Extract config information from the source in parallel with the
	 * compile.  This should be a win even on uni-processor because the
	 * config extract and the compile will share the source in memory.
	 */
	if (is_compile_type()) {
		DB_TOKEN token;
		DB_REC_SOURCE *s;
		char name[l_src + 2];
		name[0] = '/';
		memcpy(name + 1, option_src, l_src + 1);
		token =
		    lookup_db_byname(db_main, name, DB_RECTYPE_SOURCE,
				     option_srctree_n);
		assert(!DB_IS_NULL_TOKEN(token));
		s = read_db_source(db_main, token);
		check_config(db_main, token, s);
	}
	/* If the mainline has already finished then there is no point in
	 * closing and reopening the database.  One select() call saves lots
	 * of other syscalls and I/O.
	 */
	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);
	memset(&tv, 0, sizeof (tv));
	close_it = select(fd + 1, &rfds, NULL, NULL, &tv) <= 0;
	if (close_it) {
		close_db_main(db_main);
		db_main = NULL;
	}
	/* Wait until the mainline task tells me the command status */
	while ((l = read(fd, &ok, 1)) == -1 && errno == EINTR) {
	}
	if (!db_main) {
		db_main = open_db_main(1);
	}
	if (!db_main) {
		VERBOSE(2, "  do_dependencies open_db_main failed\n");
		return (1);
	}
	if (!signalled && l == 1 && ok == 'Y') {
		ret = write_depend(db_main, command, depend);
	} else {
		ret = 1;
	}
	close_db_main(db_main);
	VERBOSE(2, "  do_dependencies ret %d\n", ret);
	return (ret);
}

/* Wait for a specific task to end. */

static int
wait_for(pid_t pid, const char *text)
{
	int status;
	if (pid < 0) {
		return (0);
	}
	if (waitpid(pid, &status, 0) < 0) {
		if (debug && errno == ECHILD) {
			/* Debugging children can cause ECHILD */
			return (0);
		}
		fprintf(stderr, "%s: wait_for(%d) failed, %s: %m\n", program,
			pid, text);
		return (1);
	} else if (!WIFEXITED(status)) {
		fprintf(stderr, "%s: %s (pid %d) did not exit normally",
			program, text, pid);
		if (WIFSIGNALED(status)) {
			fprintf(stderr, " due to signal %d", WTERMSIG(status));
		}
		fprintf(stderr, "\n");
		return (1);
	} else if (WEXITSTATUS(status)) {
		fprintf(stderr, "%s: %s (pid %d) returned %d\n", program, text,
			pid, WEXITSTATUS(status));
		return (1);
	}
	return (0);
}

/* Main processing code.  See comments at the start. */

static int
process(void)
{
	int ret = 0, l;
	int in_child, out_child, close_child, close_parent, config_fd =
	    -1, null_fd = -1;
	const char *command = NULL;
	char buf[4096 * 10];
	char depend[l_target + 1 + 2 + 1];
	static int modversions = 0;	/* TBD */
	pid_t pid0 = -1, pid1 = -1, pid2 = -1;
	int pipe0[2];

	modversions = modversions && type == TYPE_CC;

	VERBOSE(1, "# %s commands\n", program);
	snprintf(depend, sizeof (depend), "%s.d", option_target2);
	while (strchr(depend, ',')) {
		*(strchr(depend, ',')) = ':';
	}
	VERBOSE(1, "  rm %s\n", depend + 1);
	if (unlink(depend + 1) && errno != ENOENT) {
		fprintf(stderr, "%s: Cannot unlink %s: %m\n", program,
			depend + 1);
		ret = 1;
	}

	if ((null_fd = open("/dev/null", O_RDONLY)) < 0) {
		fprintf(stderr, "%s: open(/dev/null) failed: %m\n", program);
		ret = 1;
	}

	switch (type) {
	case TYPE_CC:
		if (modversions) {
			command = readenv("CPP");
		} else {
			command = readenv("CC");
		}
		break;
	case TYPE_HOSTCC:
		command = readenv("HOSTCC");
		break;
	case TYPE_CPP:
		command = readenv("CPP");
		break;
	case TYPE_AS:
		command = readenv("CC");
		break;
	case TYPE_HOSTAS:
		command = readenv("HOSTCC");
		break;
	case TYPE_LD:		/* drop through */
	case TYPE_HOSTLD:	/* drop through */
	case TYPE_AR:		/* drop through */
	case TYPE_HOSTAR:
		command = readenv(option_type);
		break;
	case TYPE_USER:
		command = option_command;
		break;
	}
	if (ret == 0 && !command) {
		fprintf(stderr, "%s: No command string\n", program);
		ret = 1;
	}

	in_child = null_fd;
	out_child = -1;
	close_child = -1;
	close_parent = -1;
	if (ret == 0 && modversions) {
		if (pipe(pipe0)) {
			fprintf(stderr, "%s: pipe() failed: %m\n", program);
			ret = 1;
		}
		out_child = pipe0[1];
		close_child = pipe0[0];
		close_parent = pipe0[1];
		VERBOSE(2, "pipe0[%d,%d]\n", pipe0[0], pipe0[1]);
	}
	VERBOSE(2, "in_child %d out_child %d close_child %d close_parent %d\n",
		in_child, out_child, close_child, close_parent);

	if (ret == 0 && is_compile_type()) {
		if ((pid0 =
		     fork_command(in_child, out_child, close_child,
				  close_parent, command, "-Wp,-MD,", depend + 1,
				  " ", option_flags, " ", this_srctree,
				  option_src, " -c -o ",
				  modversions ? "-" : option_target,
				  NULL)) < 0) {
			ret = 1;
		}
		if (modversions) {
			VERBOSE(1, " |\n");
		} else {
			VERBOSE(1, " &&\n    pp_makefile5b\n");
		}
	}

	if (ret == 0 && is_link_type()) {
		if (type == TYPE_LD || type == TYPE_HOSTLD) {
			pid0 =
			    fork_command(in_child, out_child, close_child,
					 close_parent, command, option_flags,
					 " ", option_objects, " -o ",
					 option_target, NULL);
		} else {
			pid0 =
			    fork_command(in_child, out_child, close_child,
					 close_parent, command, option_flags,
					 " ", option_target, " ",
					 option_objects, NULL);
		}
		if (pid0 < 0) {
			ret = 1;
		}
		VERBOSE(1, " &&\n    pp_makefile5c\n");
	}

	if (ret == 0 && is_user_type()) {
		command = option_command;
		if ((pid0 =
		     fork_command(in_child, out_child, close_child,
				  close_parent, option_command, NULL)) < 0) {
			ret = 1;
		}
		VERBOSE(1, " &&\n    pp_makefile5d\n");
	}

	if (ret == 0) {
		/* pp_makefile5a, pp_makefile5b, extract CONFIG_xxx from the source,
		 *   standardize the dependency data.
		 * All: append the command.
		 * To minimize elapsed time, fork and do dependencies in parallel with
		 * the mainline task.
		 */
		int file_des[2];
		if (pipe(file_des)) {
			fprintf(stderr, "%s: pipe failed: %m", program);
			ret = 1;
		}
		if (ret == 0) {
			if ((pid2 = fork()) == 0) {
				if (debug) {
					sleep(10);	/* time for gdb connect */
				}
				close(file_des[1]);
				return (do_dependencies
					(file_des[0], command, depend));
			} else if (pid2 < 0) {
				fprintf(stderr, "%s: fork failed %m\n",
					program);
				ret = 1;
			} else {
				if (debug) {
					fprintf(stderr,
						"%s: process child pid %d\n",
						program, pid2);
				}
				close(file_des[0]);
				config_fd = file_des[1];
			}
		}
	}

	if (ret == 0 && modversions) {
		int pipe1[2];
		VERBOSE(1, "    pp_makefile5a |\n");
		if (pipe(pipe1)) {
			fprintf(stderr, "%s: pipe() failed: %m\n", program);
			ret = 1;
		}
		out_child = -1;
		in_child = pipe1[0];
		close_child = pipe1[1];
		close_parent = pipe1[0];
		VERBOSE(2,
			"pipe1[%d,%d] in_child %d out_child %d close_child %d close_parent %d\n",
			pipe1[0], pipe1[1], in_child, out_child, close_child,
			close_parent);
		if (!(command = readenv("CC"))) {
			ret = 1;
		}
		if (ret == 0 &&
		    (pid1 =
		     fork_command(in_child, out_child, close_child,
				  close_parent, command, option_flags, " -o ",
				  option_target, " -c -x cpp-output -",
				  NULL)) < 0) {
			ret = 1;
		}
		VERBOSE(1, "\n");
		if (ret == 0) {
			/* pp_makefile5a, copy CPP data from input to output */
			FILE *fin, *fout;
			if (!(fin = fdopen(pipe0[0], "r"))) {
				fprintf(stderr,
					"%s: fdopen(%d, \"r\") failed: %m",
					program, pipe0[0]);
				ret = 1;
			}
			if (!(fout = fdopen(pipe1[1], "w"))) {
				fprintf(stderr,
					"%s: fdopen(%d, \"w\") failed: %m",
					program, pipe1[1]);
				ret = 1;
			}
			setvbuf(fout, NULL, _IONBF, 0);
			while (ret == 0
			       && (l = fread(buf, 1, sizeof (buf), fin)) > 0) {
				if (fwrite(buf, 1, l, fout) != l) {
					fprintf(stderr,
						"%s: fwrite() cpp output failed: %m",
						program);
					ret = 1;
				}
			}
			if (ferror(fin)) {
				fprintf(stderr,
					"%s: fread() cpp input failed: %m",
					program);
				ret = 1;
			}
			if (fin && fclose(fin) && !ret) {
				fprintf(stderr,
					"%s: fclose() cpp input failed: %m",
					program);
				ret = 1;
			}
			if (fout && fclose(fout) && !ret) {
				fprintf(stderr,
					"%s: fclose() cpp output failed: %m",
					program);
				ret = 1;
			}
		}
	}

	ret = ret || wait_for(pid0, "Main command");
	ret = ret || wait_for(pid1, "modversions pipe");

	/* Tell the dependency subtask whether it can update depend.  Do not touch
	 * depend after this point, the subtask might still be writing to it.  This
	 * must be after all subtasks other than dependency have completed so we know
	 * what the final command status is.
	 */
	if (config_fd >= 0) {
		char ok = ((ret == 0 && errors == 0) ? 'Y' : 'N');
		write(config_fd, &ok, 1);
		close(config_fd);
	}
	ret = ret || wait_for(pid2, "dependencies");

	if (ret || errors) {
		unlink(option_target);
	}
	if (option_target2) {
		free(option_target2);
	}

	return (ret ? ret : errors);
}

static void
handler(int signum)
{
	VERBOSE(2, "  signalled %d\n", signum);
	if (++signalled > 2) {
		exit(2);
	}
}

int
main(int argc, char **argv)
{
	const char *writable = readenv("KBUILD_WRITABLE");
	if (!writable || *writable != 'y') {
		fprintf(stderr, "%s: KBUILD_WRITABLE is false, giving up\n",
			program);
		return (1);
	}

	signal(SIGHUP, handler);
	signal(SIGINT, handler);
	signal(SIGQUIT, handler);
	signal(SIGPIPE, handler);
	signal(SIGTERM, handler);
	signal(SIGCHLD, SIG_DFL);
	if (pp_getenv() || parse(argc, argv)) {
		return (1);
	}
	VERBOSE(1, "%s verbose %d\n", program, verbose);
	if (signalled || process()) {
		return (1);
	}
	pp_freeenv();
	return (errors);
}
